"""Ophis Intermediate Representation

    Classes for representing the Intermediate nodes upon which the
    assembler passes operate."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import Ophis.Errors as Err


class Node(object):
    """The default IR Node
    Instances of Node always have the three fields ppt(Program Point),
    nodetype(a string), and data (a list)."""
    def __init__(self, ppt, nodetype, *data):
        self.ppt = ppt
        self.nodetype = nodetype
        self.data = list(data)

    def accept(self, asmpass, env=None):
        """Implements the Visitor pattern for an assembler pass.
        Calls the routine 'asmpass.visitTYPE(self, env)' where
        TYPE is the value of self.nodetype."""
        Err.currentpoint = self.ppt
        routine = getattr(asmpass, "visit" + self.nodetype,
                          asmpass.visitUnknown)
        routine(self, env)

    def __str__(self):
        if self.nodetype != "SEQUENCE":
            return str(self.ppt) + ": " + self.nodetype + " - " + \
                   " ".join(map(str, self.data))
        else:
            return "\n".join(map(str, self.data))

    def __repr__(self):
        args = [self.ppt, self.nodetype] + self.data
        return "Node(" + ", ".join(map(repr, args)) + ")"


NullNode = Node("<none>", "None")


def SequenceNode(ppt, nodelist):
    return Node(ppt, "SEQUENCE", *nodelist)


class Expr(object):
    """Base class for Ophis expressions
    All expressions have a field called "data" and a boolean field
    called "hardcoded".  An expression is hardcoded if it has no
    symbolic values in it."""
    def __init__(self, data):
        self.data = data
        self.hardcoded = False

    def __str__(self):
        return "<UNKNOWN: " + repr(self.data) + ">"

    def valid(self, env=None, PCvalid=False):
        """Returns true if the the expression can be successfully
        evaluated in the specified environment."""
        return False

    def value(self, env=None):
        "Evaluates this expression in the given environment."
        return None


class ConstantExpr(Expr):
    "Represents a numeric constant"
    def __init__(self, data):
        self.data = data
        self.hardcoded = True

    def __str__(self):
        return str(self.data)

    def valid(self, env=None, PCvalid=False):
        return True

    def value(self, env=None):
        return self.data


class LabelExpr(Expr):
    "Represents a symbolic constant"
    def __init__(self, data):
        self.data = data
        self.hardcoded = False

    def __str__(self):
        return self.data

    def valid(self, env=None, PCvalid=False):
        return (env is not None) and self.data in env

    def value(self, env=None):
        return env[self.data]


class PCExpr(Expr):
    "Represents the current program counter: ^"
    def __init__(self):
        self.hardcoded = False

    def __str__(self):
        return "^"

    def valid(self, env=None, PCvalid=False):
        return env is not None and PCvalid

    def value(self, env=None):
        return env.getPC()


class HighByteExpr(Expr):
    "Represents the expression >{data}"
    def __init__(self, data):
        self.data = data
        self.hardcoded = data.hardcoded

    def __str__(self):
        return ">" + str(self.data)

    def valid(self, env=None, PCvalid=False):
        return self.data.valid(env, PCvalid)

    def value(self, env=None):
        val = self.data.value(env)
        return (val >> 8) & 0xff


class LowByteExpr(Expr):
    "Represents the expression <{data}"
    def __init__(self, data):
        self.data = data
        self.hardcoded = data.hardcoded

    def __str__(self):
        return "<" + str(self.data)

    def valid(self, env=None, PCvalid=False):
        return self.data.valid(env, PCvalid)

    def value(self, env=None):
        val = self.data.value(env)
        return val & 0xff


class SequenceExpr(Expr):
    """Represents an interleaving of operands (of type Expr) and
    operators (of type String).  Subclasses must provide a routine
    operate(self, firstarg, op, secondarg) that evaluates the
    operator."""
    def __init__(self, data):
        """Constructor for Sequence Expressions.  Results will be
        screwy if the data inpot isn't a list with types
        [Expr, str, Expr, str, Expr, str, ... Expr, str, Expr]."""
        self.data = data
        self.operands = [x for x in data if isinstance(x, Expr)]
        self.operators = [x for x in data if type(x) == str]
        for i in self.operands:
            if not i.hardcoded:
                self.hardcoded = False
                break
        else:
            self.hardcoded = True

    def __str__(self):
        return "[" + " ".join(map(str, self.data)) + "]"

    def valid(self, env=None, PCvalid=False):
        for i in self.operands:
            if not i.valid(env, PCvalid):
                return False
        return True

    def value(self, env=None):
        subs = list(map((lambda x: x.value(env)), self.operands))
        result = subs[0]
        index = 1
        for op in self.operators:
            result = self.operate(result, op, subs[index])
            index += 1
        return result

    def operate(self, start, op, other):
        if op == "*":
            return start * other
        if op == "/":
            return start // other
        if op == "+":
            return start + other
        if op == "-":
            return start - other
        if op == "&":
            return start & other
        if op == "|":
            return start | other
        if op == "^":
            return start ^ other



"""Core pragmas

    Provides the core assembler directives."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import Ophis.CmdLine
import Ophis.IR as IR
import Ophis.Frontend as FE
import Ophis.Errors as Err
import math
import os.path

basecharmap = "".join([chr(x) for x in range(256)])
currentcharmap = basecharmap


def reset():
    global currentcharmap, basecharmap
    FE.loadedfiles = {}
    currentcharmap = basecharmap


def pragmaOutfile(ppt, line, result):
    "Sets the output file if it hasn't already been set"
    filename = line.expect("STRING").value
    line.expect("EOL")
    if type(filename) == str and Ophis.CmdLine.outfile is None:
        Ophis.CmdLine.outfile = filename


def pragmaListfile(ppt, line, result):
    "Sets the listing file if it hasn't already been set"
    filename = line.expect("STRING").value
    line.expect("EOL")
    if type(filename) == str and Ophis.CmdLine.listfile is None:
        Ophis.CmdLine.listfile = filename


def pragmaInclude(ppt, line, result):
    "Includes a source file"
    filename = line.expect("STRING").value
    line.expect("EOL")
    if type(filename) == str:
        result.append(FE.parse_file(ppt, filename))


def pragmaRequire(ppt, line, result):
    "Includes a source file at most one time"
    filename = line.expect("STRING").value
    line.expect("EOL")
    if type(filename) == str:
        result.append(FE.parse_file(ppt, filename, True))


def pragmaIncbin(ppt, line, result):
    "Includes a binary file"
    filename = line.expect("STRING").value
    offset = IR.ConstantExpr(0)
    size = None
    if str(line.lookahead(0)) == ",":
        line.pop()
        offset = FE.parse_expr(line)
        if str(line.lookahead(0)) == ",":
            line.pop()
            size = FE.parse_expr(line)
    line.expect("EOL")
    if type(filename) == str:
        try:
            f = open(os.path.join(FE.context_directory, filename), "rb")
            if offset.hardcoded and (size is None or size.hardcoded):
                # We know how big it will be, we can just use the values.
                # First check to make sure they're sane
                if offset.value() < 0:
                    Err.log("Offset may not be negative")
                    f.close()
                    return
                f.seek(0, 2)  # Seek to end of file
                if offset.value() > f.tell():
                    Err.log("Offset runs past end of file")
                    f.close()
                    return
                if size is not None:
                    if size.value() < 0:
                        Err.log("Length may not be negative")
                        f.close()
                        return
                    if offset.value() + size.value() > f.tell():
                        Err.log(".incbin length too long")
                        f.close()
                        return
                if size is None:
                    size = IR.ConstantExpr(-1)
                f.seek(offset.value())
                bytes = f.read(size.value())
                bytes = [IR.ConstantExpr(x) for x in bytes]
                result.append(IR.Node(ppt, "Byte", *bytes))
            else:
                # offset or length could change based on label placement.
                # This seems like an unbelievably bad idea, but since we
                # don't have constant prop it will happen for any symbolic
                # alias. Don't use symbolic aliases when extracting tiny
                # pieces out of humongous files, I guess.
                bytes = f.read()
                bytes = [IR.ConstantExpr(x) for x in bytes]
                if size is None:
                    size = IR.SequenceExpr([IR.ConstantExpr(len(bytes)),
                                            "-",
                                            offset])
                result.append(IR.Node(ppt, "ByteRange", offset, size, *bytes))
            f.close()
        except IOError:
            Err.log("Could not read " + filename)
            return


def pragmaCharmap(ppt, line, result):
    "Modify the character map."
    global currentcharmap, basecharmap
    if str(line.lookahead(0)) == "EOL":
        currentcharmap = basecharmap
    else:
        bytes = readData(line)
        try:
            base = bytes[0].data
            newsubstr = "".join([chr(x.data) for x in bytes[1:]])
            currentcharmap = currentcharmap[:base] + newsubstr + \
                             currentcharmap[base + len(newsubstr):]
            if len(currentcharmap) != 256 or base < 0 or base > 255:
                Err.log("Charmap replacement out of range")
                currentcharmap = currentcharmap[:256]
        except ValueError:
            Err.log("Illegal character in .charmap directive")


def pragmaCharmapbin(ppt, line, result):
    "Load a new character map from a file"
    global currentcharmap
    filename = line.expect("STRING").value
    line.expect("EOL")
    if type(filename) == str:
        try:
            f = open(os.path.join(FE.context_directory, filename), "rb")
            bytes = f.read()
            f.close()
        except IOError:
            Err.log("Could not read " + filename)
            return
        if len(bytes) == 256:
            currentcharmap = bytes
        else:
            Err.log("Character map " + filename + " not 256 bytes long")


def pragmaOrg(ppt, line, result):
    "Relocates the PC with no output"
    newPC = FE.parse_expr(line)
    line.expect("EOL")
    result.append(IR.Node(ppt, "SetPC", newPC))


def pragmaAdvance(ppt, line, result):
    "Outputs filler until reaching the target PC"
    newPC = FE.parse_expr(line)
    if str(line.lookahead(0)) == ",":
        line.pop()
        fillexpr = FE.parse_expr(line)
    else:
        fillexpr = IR.ConstantExpr(0)
    line.expect("EOL")
    result.append(IR.Node(ppt, "Advance", newPC, fillexpr))


def pragmaCheckpc(ppt, line, result):
    "Enforces that the PC has not exceeded a certain point"
    target = FE.parse_expr(line)
    line.expect("EOL")
    result.append(IR.Node(ppt, "CheckPC", target))


def pragmaAlias(ppt, line, result):
    "Assigns an arbitrary label"
    lbl = line.expect("LABEL").value
    target = FE.parse_expr(line)
    result.append(IR.Node(ppt, "Label", lbl, target))


def pragmaSpace(ppt, line, result):
    "Reserves space in a data segment for a variable"
    lbl = line.expect("LABEL").value
    size = line.expect("NUM").value
    line.expect("EOL")
    result.append(IR.Node(ppt, "Label", lbl, IR.PCExpr()))
    result.append(IR.Node(ppt, "SetPC",
                          IR.SequenceExpr([IR.PCExpr(), "+",
                                           IR.ConstantExpr(size)])))


def pragmaText(ppt, line, result):
    "Switches to a text segment"
    next = line.expect("LABEL", "EOL")
    if next.type == "LABEL":
        line.expect("EOL")
        segment = next.value
    else:
        segment = "*text-default*"
    result.append(IR.Node(ppt, "TextSegment", segment))


def pragmaData(ppt, line, result):
    "Switches to a data segment (no output allowed)"
    next = line.expect("LABEL", "EOL")
    if next.type == "LABEL":
        line.expect("EOL")
        segment = next.value
    else:
        segment = "*data-default*"
    result.append(IR.Node(ppt, "DataSegment", segment))


def pragmaCbmfloat(ppt, line, result):
    "Parses a string into a CBM BASIC format floating point number"
    data = []
    while True:
        try:
            v_str = line.expect("STRING").value
            v = float(v_str)
            if v == 0.0:
                data.extend([0,0,0,0,0])
            else:
                if v < 0.0:
                    sign = 128
                    v = -v
                else:
                    sign = 0
                expt = math.floor(math.log(v, 2))
                if expt >= -128 and expt <= 126:
                    mantissa = v / (2**expt)
                    m1 = (mantissa - 1.0) * 128 + sign
                    m2 = m1 * 256
                    m3 = m2 * 256
                    m4 = m3 * 256
                    data.extend([int(x) % 256 for x in [expt+129,m1,m2,m3,m4]])
                else:
                    Err.log("Floating point constant out of range")
        except ValueError:
            Err.log("Expected: floating point")
        next = line.expect(',', 'EOL').type
        if next == 'EOL':
            break
    bytes = [IR.ConstantExpr(x) for x in data]
    result.append(IR.Node(ppt, "Byte", *bytes))


def readData(line):
    "Read raw data from a comma-separated list"
    if line.lookahead(0).type == "STRING":
        data = [IR.ConstantExpr(ord(x))
                for x in line.expect("STRING").value.translate(currentcharmap)]
    else:
        data = [FE.parse_expr(line)]
    next = line.expect(',', 'EOL').type
    while next == ',':
        if line.lookahead(0).type == "STRING":
            data.extend([IR.ConstantExpr(ord(x))
                         for x in line.expect("STRING").value])
        else:
            data.append(FE.parse_expr(line))
        next = line.expect(',', 'EOL').type
    return data


def pragmaByte(ppt, line, result):
    "Raw data, a byte at a time"
    bytes = readData(line)
    result.append(IR.Node(ppt, "Byte", *bytes))


def pragmaWord(ppt, line, result):
    "Raw data, a word at a time, little-endian"
    words = readData(line)
    result.append(IR.Node(ppt, "Word", *words))


def pragmaDword(ppt, line, result):
    "Raw data, a double-word at a time, little-endian"
    dwords = readData(line)
    result.append(IR.Node(ppt, "Dword", *dwords))


def pragmaWordbe(ppt, line, result):
    "Raw data, a word at a time, big-endian"
    words = readData(line)
    result.append(IR.Node(ppt, "WordBE", *words))


def pragmaDwordbe(ppt, line, result):
    "Raw data, a dword at a time, big-endian"
    dwords = readData(line)
    result.append(IR.Node(ppt, "DwordBE", *dwords))


def pragmaScope(ppt, line, result):
    "Create a new lexical scoping block"
    line.expect("EOL")
    result.append(IR.Node(ppt, "ScopeBegin"))


def pragmaScend(ppt, line, result):
    "End the innermost lexical scoping block"
    line.expect("EOL")
    result.append(IR.Node(ppt, "ScopeEnd"))


def pragmaMacro(ppt, line, result):
    "Begin a macro definition"
    lbl = line.expect("LABEL").value
    line.expect("EOL")
    result.append(IR.Node(ppt, "MacroBegin", lbl))


def pragmaMacend(ppt, line, result):
    "End a macro definition"
    line.expect("EOL")
    result.append(IR.Node(ppt, "MacroEnd"))


def pragmaInvoke(ppt, line, result):
    macro = line.expect("LABEL").value
    if line.lookahead(0).type == "EOL":
        args = []
    else:
        args = readData(line)
    result.append(IR.Node(ppt, "MacroInvoke", macro, *args))



"""Symbol tables and environments for Ophis.

   Implements the symbol lookup, through nested environments -
   any non-temporary variable is stored at the top level."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import Ophis.Errors as Err


class Environment(object):
    """Environment class.
    Controls the various scopes and global abstract execution variables."""
    def __init__(self):
        self.dicts = [{}]
        self.stack = [0]
        self.pc = 0
        self.segmentdict = {}
        self.segment = "*text-default*"
        self.scopecount = 0

    def __contains__(self, item):
        if item[0] == '_':
            for dict in [self.dicts[i] for i in self.stack]:
                if item in dict:
                    return True
            return False
        return item in self.dicts[0]

    def __getitem__(self, item):
        if item[0] == '_':
            for dict in [self.dicts[i] for i in self.stack]:
                if item in dict:
                    return dict[item]
        else:
            if item in self.dicts[0]:
                return self.dicts[0][item]
        Err.log("Unknown label '%s'" % item)
        return 0

    def __setitem__(self, item, value):
        if item[0] == '_':
            self.dicts[self.stack[0]][item] = value
        else:
            self.dicts[0][item] = value

    def __str__(self):
        return str(self.dicts)

    def getPC(self):
        return self.pc

    def setPC(self, value):
        self.pc = value

    def incPC(self, amount):
        self.pc += amount

    def getsegment(self):
        return self.segment

    def setsegment(self, segment):
        self.segmentdict[self.segment] = self.pc
        self.segment = segment
        self.pc = self.segmentdict.get(segment, 0)

    def reset(self):
        "Clears out program counter, segment, and scoping information"
        self.pc = 0
        self.segmentdict = {}
        self.segment = "*text-default*"
        self.scopecount = 0
        if len(self.stack) > 1:
            Err.log("Unmatched .scope")
        self.stack = [0]

    def newscope(self):
        "Enters a new scope for temporary labels."
        self.scopecount += 1
        self.stack.insert(0, self.scopecount)
        if len(self.dicts) <= self.scopecount:
            self.dicts.append({})

    def endscope(self):
        "Leaves a scope."
        if len(self.stack) == 1:
            Err.log("Unmatched .scend")
        self.stack.pop(0)



"Ophis - a cross-assembler for the 6502 series of chips"

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.



"""The Ophis program lister

    When displaying an assembled binary for human inspection, it is
    traditional to mix binary with reconstructed instructions that
    have all arguments precomputed. This class manages that."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import sys


class Listing(object):
    """Encapsulates the program listing. Accepts fully formatted
    instruction strings, or batches of data bytes. Batches of data
    bytes are assumed to be contiguous unless a divider is explicitly
    requested."""

    def __init__(self, fname):
        self.listing = [(0, [])]
        self.filename = fname

    def listInstruction(self, inst):
        "Add a preformatted instruction list to the listing."
        self.listing.append(inst)

    def listDivider(self, newpc):
        "Indicate that the next data block will begin at the given PC."
        self.listing.append((newpc, []))

    def listData(self, vals, pc):
        """Add a batch of data to the listing. If this starts a new
        batch of data, begin that batch at the listed PC."""
        if type(self.listing[-1]) is not tuple:
            self.listing.append((pc, []))
        self.listing[-1][1].extend(vals)

    def dump(self):
        if self.filename == "-":
            out = sys.stdout
        else:
            out = open(self.filename, "wt")
        for x in self.listing:
            if type(x) is str:
                print(x, file=out)
            elif type(x) is tuple:
                i = 0
                pc = x[0]
                while True:
                    row = x[1][i:i + 16]
                    if row == []:
                        break
                    dataline = " %04X " % (pc + i)
                    dataline += (" %02X" * len(row)) % tuple(row)
                    charline = ""
                    for c in row:
                        if c < 32 or c > 126:
                            charline += "."
                        else:
                            charline += chr(c)
                    print("%-54s  |%-16s|" % (dataline, charline), file=out)
                    i += 16
        if self.filename != "-":
            out.close()


class NullLister(object):
    "A dummy Lister that actually does nothing."
    def listInstruction(self, inst):
        pass

    def listDivider(self, newpc):
        pass

    def listData(self, vals, pc):
        pass

    def dump(self):
        pass


class LabelMapper(object):
    """Encapsulates the label map. Accepts label names, string
    representations of program points, and the location."""
    def __init__(self, fname):
        self.labeldata = []
        self.filename = fname

    def mapLabel(self, label, ppt, location):
        if label.startswith("_"):
            try:
                macroarg = int(label[1:], 10)
                # If that didn't throw, this is a macro argument
                # and we don't want to track it.
                return
            except ValueError:
                pass
            if label.startswith("_*"):
                # This is the caller side of the macro arguments,
                # and we don't want to track that either.
                return
        if label.startswith("*"):
            # Unprocess anonymous labels
            label = "*"
        shortlocs = []
        # Filenames tend to become absolute paths for better
        # error processing, but that's a disaster in these
        # charts. We split out the leafs here and then re-join
        # the macro application arrows.
        for loc in ppt.split('->'):
            shortloc = loc.split('/')[-1]
            shortloc = shortloc.split('\\')[-1]
            shortlocs.append(shortloc)
        self.labeldata.append((location, label, '->'.join(shortlocs)))

    def dump(self):
        if self.filename == "-":
            out = sys.stdout
        else:
            out = open(self.filename, "wt")
        maxlabellen = 0
        self.labeldata.sort()
        for (loc, label, srcloc) in self.labeldata:
            if len(label) > maxlabellen:
                maxlabellen = len(label)
        formatstr = "$%%04X | %%-%ds | %%s\n" % (maxlabellen)
        for l in self.labeldata:
            out.write(formatstr % l)
        if self.filename != "-":
            out.close()


class NullLabelMapper(object):
    "A dummy LabelMapper that actually does nothing."
    def mapLabel(self, label, ppt, location):
        pass

    def dump(self):
        pass



"""Error logging

    Keeps track of the number of errors inflicted so far, and
    where in the assembly the errors are occurring."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import sys

count = 0
currentpoint = "<Top Level>"


def log(err):
    """Reports an error at the current program point, and increases
the global error count."""
    global count
    count = count + 1
    print(currentpoint + ": " + err, file=sys.stderr)


def report():
    "Print out the number of errors."
    if count == 0:
        print("No errors", file=sys.stderr)
    elif count == 1:
        print("1 error", file=sys.stderr)
    else:
        print(str(count) + " errors", file=sys.stderr)



"""Lexer and Parser

    Constructs a list of IR nodes from a list of input strings."""

import Ophis.Errors as Err
import Ophis.Opcodes as Ops
import Ophis.IR as IR
import Ophis.CmdLine as Cmd
import sys
import os
import os.path

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.


loadedfiles = {}
templabelcount = 0


class Lexeme(object):
    "Class for lexer tokens.  Used by lexer and parser."
    def __init__(self, type="UNKNOWN", value=None):
        self.type = type.upper()
        self.value = value

    def __str__(self):
        if self.value is None:
            return self.type
        else:
            return self.type + ":" + str(self.value)

    def __repr__(self):
        return "Lexeme(" + repr(self.type) + ", " + repr(self.value) + ")"

    def matches(self, other):
        "1 if Lexemes a and b have the same type."
        return self.type == other.type


bases = {"$": ("hexadecimal", 16),
         "%": ("binary", 2),
         "0": ("octal", 8)}


punctuation = "#,`<>():.+-*/&|^[]"


def lex(point, line):
    """Turns a line of source into a sequence of lexemes."""
    Err.currentpoint = point
    result = []

    def is_opcode(op):
        "Tests whether a string is an opcode or an identifier"
        return op in Ops.opcodes

    def add_token(token):
        "Converts a substring into a single lexeme"
        if token == "":
            return
        if token == "0":
            result.append(Lexeme("NUM", 0))
            return
        firstchar = token[0]
        rest = token[1:]
        if firstchar == '"':
            result.append(Lexeme("STRING", rest))
            return
        elif firstchar in bases:
            try:
                result.append(Lexeme("NUM", int(rest, bases[firstchar][1])))
                return
            except ValueError:
                Err.log('Invalid ' + bases[firstchar][0] + ' constant: ' +
                        rest)
                result.append(Lexeme("NUM", 0))
                return
        elif firstchar.isdigit():
            try:
                result.append(Lexeme("NUM", int(token)))
            except ValueError:
                Err.log('Identifiers may not begin with a number')
                result.append(Lexeme("LABEL", "ERROR"))
            return
        elif firstchar == "'":
            if len(rest) == 1:
                result.append(Lexeme("NUM", ord(rest)))
            else:
                Err.log("Invalid character constant '" + rest + "'")
                result.append(Lexeme("NUM", 0))
            return
        elif firstchar in punctuation:
            if rest != "":
                Err.log("Internal lexer error!  '" + token + "' can't happen!")
            result.append(Lexeme(firstchar))
            return
        else:   # Label, opcode, or index register
            id = token.lower()
            if is_opcode(id):
                result.append(Lexeme("OPCODE", id))
            elif id == "x":
                result.append(Lexeme("X"))
            elif id == "y":
                result.append(Lexeme("Y"))
            elif id == "z":
                result.append(Lexeme("Z"))
            elif id == "sp":
                result.append(Lexeme("SP"))
            else:
                result.append(Lexeme("LABEL", id))
            return
        # should never reach here
        Err.log("Internal lexer error: add_token fall-through")

    def add_EOL():
        "Adds an end-of-line lexeme"
        result.append(Lexeme("EOL"))

    # Actual routine begins here
    value = ""
    quotemode = False
    backslashmode = False
    for c in line.strip():
        if backslashmode:
            backslashmode = False
            value = value + c
        elif c == "\\":
            backslashmode = True
        elif quotemode:
            if c == '"':
                quotemode = False
            else:
                value = value + c
        elif c == ';':
            add_token(value)
            value = ""
            break
        elif c == '.' and value != "":
            value = value + c
        elif c.isspace():
            add_token(value)
            value = ""
        elif c in punctuation:
            add_token(value)
            add_token(c)
            value = ""
        elif c == '"':
            add_token(value)
            value = '"'
            quotemode = True
        else:
            value = value + c
    if backslashmode:
        Err.log("Backslashed newline")
    if quotemode:
        Err.log("Unterminated string constant")
    add_token(value)
    add_EOL()
    return result


class ParseLine(object):
    "Maintains the parse state of a line of code. Enables arbitrary lookahead."

    def __init__(self, lexemes):
        self.lexemes = lexemes
        self.location = 0

    def lookahead(self, i):
        """Returns the token i units ahead in the parse.
    lookahead(0) returns the next token; trying to read off the end of
    the sequence returns the last token in the sequence (usually EOL)."""
        target = self.location + i
        if target >= len(self.lexemes):
            target = -1
        return self.lexemes[target]

    def pop(self):
        "Returns and removes the next element in the line."
        old = self.location
        if self.location < len(self.lexemes) - 1:
            self.location += 1
        return self.lexemes[old]

    def expect(self, *tokens):
        """Reads a token from the ParseLine line and returns it if it's of a
    type in the sequence tokens.  Otherwise, it logs an error."""
        token = self.pop()
        if token.type in tokens:
            return token
        if 'LABEL' in tokens:
            if token.type in ['X', 'Y', 'Z', 'SP']:
                token.value = token.type.lower()
                token.type = 'LABEL'
                return token
            elif token.type == 'OPCODE':
                token.type = 'LABEL'
                return token
        Err.log('Expected: "' + '", "'.join(tokens) + '"')
        return token


pragma_modules = []


def parse_expr(line):
    "Parses an Ophis arithmetic expression."

    def atom():
        "Parses lowest-priority expression components."
        global templabelcount
        next = line.lookahead(0).type
        if next == "NUM":
            return IR.ConstantExpr(line.expect("NUM").value)
        elif next in ["LABEL", "X", "Y", "Z", "SP", "OPCODE"]:
            return IR.LabelExpr(line.expect("LABEL").value)
        elif next == "^":
            line.expect("^")
            return IR.PCExpr()
        elif next == "[":
            line.expect("[")
            result = parse_expr(line)
            line.expect("]")
            return result
        elif next == "+":
            offset = 0
            while next == "+":
                offset += 1
                line.expect("+")
                next = line.lookahead(0).type
            return IR.LabelExpr("*" + str(templabelcount + offset))
        elif next == "-":
            offset = 1
            while next == "-":
                offset -= 1
                line.expect("-")
                next = line.lookahead(0).type
            return IR.LabelExpr("*" + str(templabelcount + offset))
        elif next == ">":
            line.expect(">")
            return IR.HighByteExpr(atom())
        elif next == "<":
            line.expect("<")
            return IR.LowByteExpr(atom())
        else:
            Err.log('Expected: expression')

    def precedence_read(constructor, reader, separators):
        """Handles precedence.  The reader argument is a function that returns
    expressions that bind more tightly than these; separators is a list
    of strings naming the operators at this precedence level.  The
    constructor argument is a class, indicating what node type holds
    objects of this precedence level.

    Returns a list of Expr objects with separator strings between them."""
        result = [reader()]  # first object
        nextop = line.lookahead(0).type
        while (nextop in separators):
            line.expect(nextop)
            result.append(nextop)
            result.append(reader())
            nextop = line.lookahead(0).type
        if len(result) == 1:
            return result[0]
        return constructor(result)

    def term():
        "Parses * and /"
        return precedence_read(IR.SequenceExpr, atom, ["*", "/"])

    def arith():
        "Parses + and -"
        return precedence_read(IR.SequenceExpr, term, ["+", "-"])

    def bits():
        "Parses &, |, and ^"
        return precedence_read(IR.SequenceExpr, arith, ["&", "|", "^"])

    return bits()


def parse_line(ppt, lexemelist):
    "Turn a line of source into an IR Node."
    Err.currentpoint = ppt
    result = []
    line = ParseLine(lexemelist)

    def aux():
        "Accumulates all IR nodes defined by this line."
        if line.lookahead(0).type == "EOL":
            pass
        elif line.lookahead(1).type == ":":
            newlabel = line.expect("LABEL").value
            line.expect(":")
            result.append(IR.Node(ppt, "Label", newlabel, IR.PCExpr()))
            aux()
        elif line.lookahead(0).type == "*":
            global templabelcount
            templabelcount = templabelcount + 1
            result.append(IR.Node(ppt, "Label", "*" + str(templabelcount),
                                  IR.PCExpr()))
            line.expect("*")
            aux()
        elif line.lookahead(0).type == "." or line.lookahead(0).type == "`":
            which = line.expect(".", "`").type
            if (which == "."):
                pragma = line.expect("LABEL").value
            else:
                pragma = "invoke"
            pragmaFunction = "pragma" + pragma.title()
            for mod in pragma_modules:
                if hasattr(mod, pragmaFunction):
                    getattr(mod, pragmaFunction)(ppt, line, result)
                    break
            else:
                Err.log("Unknown pragma " + pragma)
        else:   # Instruction
            opcode = line.expect("OPCODE").value
            arg2 = None
            if line.lookahead(0).type == "#":
                mode = "Immediate"
                line.expect("#")
                arg = parse_expr(line)
                line.expect("EOL")
            elif line.lookahead(0).type == "(":
                line.expect("(")
                arg = parse_expr(line)
                if line.lookahead(0).type == ",":
                    line.expect(",")
                    if line.lookahead(0).type == "X":
                        mode = "PointerX"
                        line.expect("X")
                        line.expect(")")
                        line.expect("EOL")
                    else:
                        mode = "PointerSPY"
                        line.expect("SP")
                        line.expect(")")
                        line.expect(",")
                        line.expect("Y")
                        line.expect("EOL")
                else:
                    line.expect(")")
                    tok = line.expect(",", "EOL").type
                    if tok == "EOL":
                        mode = "Pointer"
                    else:
                        if line.lookahead(0).type == "Y":
                            mode = "PointerY"
                            line.expect("Y")
                            line.expect("EOL")
                        else:
                            mode = "PointerZ"
                            line.expect("Z")
                            line.expect("EOL")
            elif line.lookahead(0).type == "EOL":
                mode = "Implied"
                arg = None
            else:
                arg = parse_expr(line)
                tok = line.expect("EOL", ",").type
                if tok == ",":
                    # Parser has to special-case the BBXn instructions,
                    # Which uniquely take two addresses
                    if opcode[:3] in ["bbs", "bbr"]:
                        arg2 = parse_expr(line)
                        mode = "Memory2"
                    else:
                        tok = line.expect("X", "Y", "Z").type
                        if tok == "X":
                            mode = "MemoryX"
                        elif tok == "Y":
                            mode = "MemoryY"
                        else:
                            mode = "MemoryZ"
                    line.expect("EOL")
                else:
                    mode = "Memory"
            result.append(IR.Node(ppt, mode, opcode, arg, arg2))

    aux()
    result = [node for node in result if node is not IR.NullNode]
    if len(result) == 0:
        return IR.NullNode
    if len(result) == 1:
        return result[0]
    return IR.SequenceNode(ppt, result)


context_directory = None


def parse_file(ppt, filename, load_once=False):
    "Loads an Ophis source file, and returns an IR list."
    global context_directory, loadedfiles
    Err.currentpoint = ppt
    old_context = context_directory
    if filename != '-':
        if context_directory is not None:
            filename = os.path.abspath(os.path.join(context_directory,
                                                    filename))
        if load_once and filename in loadedfiles:
            if Cmd.print_loaded_files:
                print("Skipping " + filename, file=sys.stderr)
            return IR.NullNode
        loadedfiles[filename] = True
    if Cmd.print_loaded_files:
        if filename != '-':
            print("Loading " + filename, file=sys.stderr)
        else:
            print("Loading from standard input", file=sys.stderr)
    try:
        if filename != '-':
            if context_directory is not None:
                filename = os.path.join(context_directory, filename)
            f = open(filename, "rt")
            linelist = f.readlines()
            f.close()
            context_directory = os.path.abspath(os.path.dirname(filename))
        else:
            context_directory = os.getcwd()
            linelist = sys.stdin.readlines()
        pptlist = ["%s:%d" % (filename, i + 1) for i in range(len(linelist))]
        lexlist = list(map(lex, pptlist, linelist))
        IRlist = list(map(parse_line, pptlist, lexlist))
        IRlist = [node for node in IRlist if node is not IR.NullNode]
        context_directory = old_context
        return IR.SequenceNode(ppt, IRlist)
    except IOError:
        Err.log("Could not read " + filename)
        context_directory = old_context
        return IR.NullNode


def parse(filenames):
    """Top level parsing routine, taking a source file name
    list and returning an IR list."""
    global templabelcount
    templabelcount = 0
    nodes = [parse_file("<Top Level>", f) for f in filenames]
    if len(nodes) == 1:
        return nodes[0]
    return IR.SequenceNode("<Top level>", nodes)



"""Macro support for Ophis.

    Ophis Macros are cached SequenceNodes with arguments
    set via .alias commands and prevented from escaping
    with .scope and .scend commands."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import sys

import Ophis.IR as IR
import Ophis.Errors as Err

macros = {}
currentname = None
currentbody = None


def newMacro(name):
    "Start creating a new macro with the specified name."
    global currentname
    global currentbody
    global macros
    if currentname is not None:
        Err.log("Internal error!  Nested macro attempt!")
    else:
        if name in macros:
            Err.log("Duplicate macro definition '%s'" % name)
        currentname = name
        currentbody = []


def registerNode(node):
    global currentbody
    currentbody.append(IR.Node(node.ppt, node.nodetype, *node.data))


def endMacro():
    global currentname
    global currentbody
    global macros
    if currentname is None:
        Err.log("Internal error!  Ended a non-existent macro!")
    else:
        macros[currentname] = currentbody
        currentname = None
        currentbody = None


def expandMacro(ppt, name, arglist):
    global macros
    if name not in macros:
        Err.log("Undefined macro '%s'" % name)
        return IR.NullNode
    argexprs = [IR.Node(ppt, "Label", "_*%d" % i, arg)
                for (i, arg) in zip(range(1, sys.maxsize), arglist)]
    bindexprs = [IR.Node(ppt, "Label", "_%d" % i, IR.LabelExpr("_*%d" % i))
                 for i in range(1, len(arglist) + 1)]
    body = [IR.Node("%s->%s" % (ppt, node.ppt), node.nodetype, *node.data)
            for node in macros[name]]
    invocation = [IR.Node(ppt, "ScopeBegin")] + argexprs + \
                 [IR.Node(ppt, "ScopeBegin")] + bindexprs + body + \
                 [IR.Node(ppt, "ScopeEnd"), IR.Node(ppt, "ScopeEnd")]
    return IR.SequenceNode(ppt, invocation)


def dump():
    global macros
    for mac in macros:
        body = macros[mac]
        print("Macro: " + mac, file=sys.stderr)
        for node in body:
            print(node, file=sys.stderr)
        print("", file=sys.stderr)



"""The Ophis Assembler passes

    Ophis's design philosophy is to build the IR once, then run a great
    many assembler passes over the result.  Thus, each pass does a
    single, specialized job.  When strung together, the full
    translation occurs.  This structure also makes the assembler
    very extensible; additional analyses or optimizations may be
    added as new subclasses of Pass."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import sys
import Ophis.Errors as Err
import Ophis.IR as IR
import Ophis.Opcodes as Ops
import Ophis.CmdLine as Cmd
import Ophis.Listing as Listing
import Ophis.Macro as Macro


class Pass(object):
    """Superclass for all assembler passes.  Automatically handles IR
    types that modify the environent's structure, and by default
    raises an error on anything else.  Override visitUnknown in your
    extension pass to produce a pass that accepts everything."""
    name = "Default Pass"

    def __init__(self):
        self.writeOK = True

    def visitNone(self, node, env):
        pass

    def visitSEQUENCE(self, node, env):
        Err.currentpoint = node.ppt
        for n in node.data:
            n.accept(self, env)

    def visitDataSegment(self, node, env):
        self.writeOK = False
        env.setsegment(node.data[0])

    def visitTextSegment(self, node, env):
        self.writeOK = True
        env.setsegment(node.data[0])

    def visitScopeBegin(self, node, env):
        env.newscope()

    def visitScopeEnd(self, node, env):
        env.endscope()

    def visitUnknown(self, node, env):
        Err.log("Internal error!  " + self.name +
                " cannot understand node type " + node.nodetype)

    def prePass(self):
        pass

    def postPass(self):
        pass

    def go(self, node, env):
        """Prepares the environment and runs this pass, possibly
        printing debugging information."""
        if Err.count == 0:
            if Cmd.print_pass:
                print("Running: " + self.name, file=sys.stderr)
            env.reset()
            self.prePass()
            node.accept(self, env)
            self.postPass()
            env.reset()
            if Cmd.print_labels:
                print("Current labels:", file=sys.stderr)
                print(env, file=sys.stderr)
            if Cmd.print_ir:
                print("Current IR:", file=sys.stderr)
                print(node, file=sys.stderr)


class FixPoint(object):
    """A specialized class that is not a pass but can be run like one.
    This class takes a list of passes and a "fixpoint" function."""
    def __init__(self, name, passes, fixpoint):
        self.name = name
        self.passes = passes
        self.fixpoint = fixpoint

    def go(self, node, env):
        """Runs this FixPoint's passes, in order, until the fixpoint
        is true.  Always runs the passes at least once."""
        for i in range(100):
            if Err.count != 0:
                break
            for p in self.passes:
                p.go(node, env)
            if Err.count != 0:
                break
            if self.fixpoint():
                break
            if Cmd.print_pass:
                print("Fixpoint failed, looping back", file=sys.stderr)
        else:
            Err.log("Can't make %s converge!  Maybe there's a recursive "
                    "dependency somewhere?" % self.name)


class DefineMacros(Pass):
    "Extract macro definitions and remove them from the IR"
    name = "Macro definition pass"

    def prePass(self):
        self.inDef = False
        self.nestedError = False

    def postPass(self):
        if self.inDef:
            Err.log("Unmatched .macro")
        elif Cmd.print_ir:
            print("Macro definitions:", file=sys.stderr)
            Macro.dump()

    def visitMacroBegin(self, node, env):
        if self.inDef:
            Err.log("Nested macro definition")
            self.nestedError = True
        else:
            Macro.newMacro(node.data[0])
            node.nodetype = "None"
            node.data = []
            self.inDef = True

    def visitMacroEnd(self, node, env):
        if self.inDef:
            Macro.endMacro()
            node.nodetype = "None"
            node.data = []
            self.inDef = False
        elif not self.nestedError:
            Err.log("Unmatched .macend")

    def visitUnknown(self, node, env):
        if self.inDef:
            Macro.registerNode(node)
            node.nodetype = "None"
            node.data = []


class ExpandMacros(Pass):
    "Replace macro invocations with the appropriate text"
    name = "Macro expansion pass"

    def prePass(self):
        self.changed = False

    def visitMacroInvoke(self, node, env):
        replacement = Macro.expandMacro(node.ppt, node.data[0], node.data[1:])
        node.nodetype = replacement.nodetype
        node.data = replacement.data
        self.changed = True

    def visitUnknown(self, node, env):
        pass


class InitLabels(Pass):
    "Finds all reachable labels"
    name = "Label initialization pass"

    def __init__(self):
        Pass.__init__(self)
        self.labelmap = {}
        self.runcount = 0

    def prePass(self):
        self.changed = False
        self.PCvalid = True
        self.runcount += 1

    def visitAdvance(self, node, env):
        self.PCvalid = node.data[0].valid(env, self.PCvalid)

    def visitSetPC(self, node, env):
        self.PCvalid = node.data[0].valid(env, self.PCvalid)

    def visitLabel(self, node, env):
        (label, val) = node.data
        fulllabel = "%d:%s" % (env.stack[0], label)
        if fulllabel in self.labelmap and self.labelmap[fulllabel] is not node:
            Err.log("Duplicate label definition '%s'" % label)
        if fulllabel not in self.labelmap:
            self.labelmap[fulllabel] = node
        if val.valid(env, self.PCvalid) and label not in env:
            env[label] = 0
            self.changed = True
        if label in ['a', 'x', 'y'] and self.runcount == 1:
            print(str(node.ppt) + ": WARNING: " \
                "using register name as label", file=sys.stderr)
        if label in Ops.opcodes and self.runcount == 1:
            print(str(node.ppt) + ": WARNING: " \
                "using opcode name as label", file=sys.stderr)

    def visitUnknown(self, node, env):
        pass


class CircularityCheck(Pass):
    "Checks for circular label dependencies"
    name = "Circularity check pass"

    def prePass(self):
        self.changed = False
        self.PCvalid = True

    def visitAdvance(self, node, env):
        PCvalid = self.PCvalid
        self.PCvalid = node.data[0].valid(env, self.PCvalid)
        if not node.data[0].valid(env, PCvalid):
            Err.log("Undefined or circular reference on .advance")

    def visitSetPC(self, node, env):
        PCvalid = self.PCvalid
        self.PCvalid = node.data[0].valid(env, self.PCvalid)
        if not node.data[0].valid(env, PCvalid):
            Err.log("Undefined or circular reference on program counter set")

    def visitCheckPC(self, node, env):
        if not node.data[0].valid(env, self.PCvalid):
            Err.log("Undefined or circular reference on program counter check")

    def visitLabel(self, node, env):
        (label, val) = node.data
        if not val.valid(env, self.PCvalid):
            Err.log("Undefined or circular dependency for label '%s'" % label)

    def visitUnknown(self, node, env):
        pass


class CheckExprs(Pass):
    "Ensures all expressions can resolve"
    name = "Expression checking pass"

    def visitUnknown(self, node, env):
        # Throw away result, just confirm validity of all expressions
        for i in [x for x in node.data if isinstance(x, IR.Expr)]:
            i.value(env)


class EasyModes(Pass):
    "Assigns address modes to hardcoded and branch instructions"
    name = "Easy addressing modes pass"

    def visitMemory(self, node, env):
        if Ops.opcodes[node.data[0]][Ops.modes.index("Relative")] is not None:
            node.nodetype = "Relative"
            return
        if Ops.opcodes[node.data[0]][Ops.modes.index("RelativeLong")] is not None:
            node.nodetype = "RelativeLong"
            return
        if node.data[1].hardcoded:
            if not collapse_no_index(node, env):
                node.nodetype = "Absolute"

    def visitMemoryX(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_x(node, env):
                node.nodetype = "AbsoluteX"

    def visitMemoryY(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_y(node, env):
                node.nodetype = "AbsoluteY"

    def visitMemory2(self, node, env):
        node.nodetype = "ZPRelative"

    def visitPointer(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_no_index_ind(node, env):
                node.nodetype = "Indirect"

    def visitPointerX(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_x_ind(node, env):
                node.nodetype = "AbsIndX"

    def visitPointerY(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_y_ind(node, env):
                node.nodetype = "AbsIndY"

    def visitPointerSPY(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_spy_ind(node, env):
                node.nodetype = "AbsIndSPY"

    def visitPointerZ(self, node, env):
        if node.data[1].hardcoded:
            if not collapse_z_ind(node, env):
                node.nodetype = "AbsIndZ"

    def visitUnknown(self, node, env):
        pass


class PCTracker(Pass):
    "Superclass for passes that need an accurate program counter."
    name = "**BUG** PC Tracker Superpass used directly"

    def visitSetPC(self, node, env):
        env.setPC(node.data[0].value(env))

    def visitAdvance(self, node, env):
        env.setPC(node.data[0].value(env))

    def visitImplied(self, node, env):
        env.incPC(1)

    def visitImmediate(self, node, env):
        env.incPC(2)

    def visitImmediateLong(self, node, env):
        env.incPC(3)

    def visitIndirectX(self, node, env):
        env.incPC(2)

    def visitIndirectY(self, node, env):
        env.incPC(2)

    def visitIndirectSPY(self, node, env):
        env.incPC(2)

    def visitIndirectZ(self, node, env):
        env.incPC(2)

    def visitZPIndirect(self, node, env):
        env.incPC(2)

    def visitZeroPage(self, node, env):
        env.incPC(2)

    def visitZeroPageX(self, node, env):
        env.incPC(2)

    def visitZeroPageY(self, node, env):
        env.incPC(2)

    def visitRelative(self, node, env):
        env.incPC(2)

    def visitRelativeLong(self, node, env):
        env.incPC(3)

    def visitZPRelative(self, node, env):
        env.incPC(3)

    def visitIndirect(self, node, env):
        env.incPC(3)

    def visitAbsolute(self, node, env):
        env.incPC(3)

    def visitAbsoluteX(self, node, env):
        env.incPC(3)

    def visitAbsoluteY(self, node, env):
        env.incPC(3)

    def visitAbsIndX(self, node, env):
        env.incPC(3)

    def visitAbsIndY(self, node, env):
        env.incPC(3)

    def visitAbsIndZ(self, node, env):
        env.incPC(3)

    def visitMemory(self, node, env):
        env.incPC(3)

    def visitMemoryX(self, node, env):
        env.incPC(3)

    def visitMemoryY(self, node, env):
        env.incPC(3)

    def visitMemoryZ(self, node, env):
        env.incPC(3)

    def visitPointer(self, node, env):
        env.incPC(3)

    def visitPointerX(self, node, env):
        env.incPC(3)

    def visitPointerY(self, node, env):
        env.incPC(3)

    def visitCheckPC(self, node, env):
        pass

    def visitLabel(self, node, env):
        pass

    def visitByte(self, node, env):
        env.incPC(len(node.data))

    def visitByteRange(self, node, env):
        if node.data[1].valid(env):
            env.incPC(node.data[1].value(env))

    def visitWord(self, node, env):
        env.incPC(len(node.data) * 2)

    def visitDword(self, node, env):
        env.incPC(len(node.data) * 4)

    def visitWordBE(self, node, env):
        env.incPC(len(node.data) * 2)

    def visitDwordBE(self, node, env):
        env.incPC(len(node.data) * 4)


class UpdateLabels(PCTracker):
    "Computes the new values for all entries in the symbol table"
    name = "Label Update Pass"

    def prePass(self):
        self.changed = False

    def visitLabel(self, node, env):
        (label, val) = node.data
        old = env[label]
        env[label] = val.value(env)
        if old != env[label]:
            self.changed = True


class Collapse(PCTracker):
    "Selects as many zero-page instructions to convert as possible."
    name = "Instruction Collapse Pass"

    def prePass(self):
        self.changed = False

    def visitMemory(self, node, env):
        self.changed |= collapse_no_index(node, env)
        PCTracker.visitMemory(self, node, env)

    def visitMemoryX(self, node, env):
        self.changed |= collapse_x(node, env)
        PCTracker.visitMemoryX(self, node, env)

    def visitMemoryY(self, node, env):
        self.changed |= collapse_y(node, env)
        PCTracker.visitMemoryY(self, node, env)

    def visitMemoryZ(self, node, env):
        PCTracker.visitMemoryZ(self, node, env)

    def visitPointer(self, node, env):
        self.changed |= collapse_no_index_ind(node, env)
        PCTracker.visitPointer(self, node, env)

    def visitPointerX(self, node, env):
        self.changed |= collapse_x_ind(node, env)
        PCTracker.visitPointerX(self, node, env)

    def visitPointerY(self, node, env):
        self.changed |= collapse_y_ind(node, env)
        PCTracker.visitPointerY(self, node, env)

    # Previously zero-paged elements may end up un-zero-paged by
    # the branch extension pass. Force them to Absolute equivalents
    # if this happens.

    def visitImmediate(self, node, env):
        if node.data[1].value(env) >= 0x100:
            if Ops.opcodes[node.data[0]][Ops.modes.index("ImmediateLong")] is not None:
                node.nodetype = "ImmediateLong"
                PCTracker.visitImmediateLong(self, node, env)
                self.changed = True
                return
        PCTracker.visitImmediate(self, node, env)

    def visitZeroPage(self, node, env):
        if node.data[1].value(env) >= 0x100:
            if Ops.opcodes[node.data[0]][Ops.modes.index("Absolute")] is not None:
                node.nodetype = "Absolute"
                PCTracker.visitAbsolute(self, node, env)
                self.changed = True
                return
        PCTracker.visitZeroPage(self, node, env)

    def visitZeroPageX(self, node, env):
        if node.data[1].value(env) >= 0x100:
            if Ops.opcodes[node.data[0]][Ops.modes.index("Absolute, X")] is not None:
                node.nodetype = "AbsoluteX"
                PCTracker.visitAbsoluteX(self, node, env)
                self.changed = True
                return
        PCTracker.visitZeroPageX(self, node, env)

    def visitZeroPageY(self, node, env):
        if node.data[1].value(env) >= 0x100:
            if Ops.opcodes[node.data[0]][Ops.modes.index("Absolute, Y")] is not None:
                node.nodetype = "AbsoluteY"
                PCTracker.visitAbsoluteY(self, node, env)
                self.changed = True
                return
        PCTracker.visitZeroPageY(self, node, env)


def collapse_no_index(node, env):
    """Transforms a Memory node into a ZeroPage one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("Zero Page")] is not None:
            node.nodetype = "ZeroPage"
            return True
    return False


def collapse_x(node, env):
    """Transforms a MemoryX node into a ZeroPageX one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("Zero Page, X")] is not None:
            node.nodetype = "ZeroPageX"
            return True
    return False


def collapse_y(node, env):
    """Transforms a MemoryY node into a ZeroPageY one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("Zero Page, Y")] is not None:
            node.nodetype = "ZeroPageY"
            return True
    return False

def collapse_no_index_ind(node, env):
    """Transforms a Pointer node into a ZPIndirect one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("(Zero Page)")] is not None:
            node.nodetype = "ZPIndirect"
            return True
    return False


def collapse_x_ind(node, env):
    """Transforms a PointerX node into an IndirectX one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("(Zero Page, X)")] is not None:
            node.nodetype = "IndirectX"
            return True
    return False


def collapse_y_ind(node, env):
    """Transforms a PointerY node into an IndirectY one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("(Zero Page), Y")] is not None:
            node.nodetype = "IndirectY"
            return True
    return False

def collapse_spy_ind(node, env):
    """Transforms a PointerSPY node into an IndirectY one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("(Zero Page, SP), Y")] is not None:
            node.nodetype = "IndirectSPY"
            return True
    return False

def collapse_z_ind(node, env):
    """Transforms a PointerZ node into an IndirectZ one if possible.
    Returns boolean indicating whether or not it made the collapse."""
    if node.data[1].value(env) < 0x100:
        if Ops.opcodes[node.data[0]][Ops.modes.index("(Zero Page), Z")] is not None:
            node.nodetype = "IndirectZ"
            return True
    return False


class ExtendBranches(PCTracker):
    """Eliminates any branch instructions that would end up going past
    the 128-byte range, and replaces them with a branch-jump pair."""
    name = "Branch Expansion Pass"
    reversed = {'bcc': 'bcs',
                'bcs': 'bcc',
                'beq': 'bne',
                'bmi': 'bpl',
                'bne': 'beq',
                'bpl': 'bmi',
                'bvc': 'bvs',
                'bvs': 'bvc',
                # 65c02 ones. 'bra' is special, though, having no inverse
                'bbr0': 'bbs0',
                'bbs0': 'bbr0',
                'bbr1': 'bbs1',
                'bbs1': 'bbr1',
                'bbr2': 'bbs2',
                'bbs2': 'bbr2',
                'bbr3': 'bbs3',
                'bbs3': 'bbr3',
                'bbr4': 'bbs4',
                'bbs4': 'bbr4',
                'bbr5': 'bbs5',
                'bbs5': 'bbr5',
                'bbr6': 'bbs6',
                'bbs6': 'bbr6',
                'bbr7': 'bbs7',
                'bbs7': 'bbr7'
                }

    def prePass(self):
        self.changed = False

    def visitRelative(self, node, env):
        (opcode, expr) = node.data[:2]
        arg = expr.value(env)
        arg = arg - (env.getPC() + 2)
        if arg < -128 or arg > 127:
            if Cmd.enable_4502_exts:
                node.nodetype = "RelativeLong"
                if Cmd.warn_on_branch_extend:
                    print(str(node.ppt) + ": WARNING: " \
                        "branch out of range, replacing with 16-bit relative branch", file=sys.stderr)
            else:
                if opcode == 'bra':
                    # If BRA - BRanch Always - is out of range, it's a JMP.
                    node.data = ('jmp', expr, None)
                    node.nodetype = "Absolute"
                    if Cmd.warn_on_branch_extend:
                        print(str(node.ppt) + ": WARNING: " \
                            "bra out of range, replacing with jmp", file=sys.stderr)
                else:
                    # Otherwise, we replace it with a 'macro' of sorts by hand:
                    # $branch LOC -> $reversed_branch ^+5; JMP LOC
                    # We don't use temp labels here because labels need to have
                    # been fixed in place by this point, and JMP is always 3
                    # bytes long.
                    expansion = [IR.Node(node.ppt, "Relative",
                                     ExtendBranches.reversed[opcode],
                                     IR.SequenceExpr([IR.PCExpr(), "+",
                                                      IR.ConstantExpr(5)]),
                                     None),
                             IR.Node(node.ppt, "Absolute", 'jmp', expr, None)]
                    node.nodetype = 'SEQUENCE'
                    node.data = expansion
                    if Cmd.warn_on_branch_extend:
                        print(str(node.ppt) + ": WARNING: " + \
                                       opcode + " out of range, " \
                                       "replacing with " + \
                                       ExtendBranches.reversed[opcode] + \
                                       "/jmp combo", file=sys.stderr)
                    self.changed = True
                    node.accept(self, env)
        else:
            PCTracker.visitRelative(self, node, env)

    def visitZPRelative(self, node, env):
        (opcode, tested, expr) = node.data
        arg = expr.value(env)
        arg = arg - (env.getPC() + 3)
        if arg < -128 or arg > 127:
            # Otherwise, we replace it with a 'macro' of sorts by hand:
            # $branch LOC -> $reversed_branch ^+6; JMP LOC
            # We don't use temp labels here because labels need to have
            # been fixed in place by this point, and JMP is always 3
            # bytes long.
            expansion = [IR.Node(node.ppt, "ZPRelative",
                                 ExtendBranches.reversed[opcode],
                                 tested,
                                 IR.SequenceExpr([IR.PCExpr(), "+",
                                                  IR.ConstantExpr(6)])),
                         IR.Node(node.ppt, "Absolute", 'jmp', expr, None)]
            node.nodetype = 'SEQUENCE'
            node.data = expansion
            if Cmd.warn_on_branch_extend:
                print(str(node.ppt) + ": WARNING: " + \
                    opcode + " out of range, " \
                    "replacing with " + \
                    ExtendBranches.reversed[opcode] + \
                    "/jmp combo", file=sys.stderr)
            self.changed = True
            node.accept(self, env)
        else:
            PCTracker.visitZPRelative(self, node, env)


class NormalizeModes(Pass):
    """Eliminates the intermediate "Memory" and "Pointer" nodes,
    converting them to "Absolute"."""
    name = "Mode Normalization pass"

    def visitMemory(self, node, env):
        node.nodetype = "Absolute"

    def visitMemoryX(self, node, env):
        node.nodetype = "AbsoluteX"

    def visitMemoryY(self, node, env):
        node.nodetype = "AbsoluteY"

    def visitPointer(self, node, env):
        node.nodetype = "Indirect"

    def visitPointerX(self, node, env):
        node.nodetype = "AbsIndX"

    # If we ever hit a PointerY by this point, we have a bug.

    def visitPointerY(self, node, env):
        node.nodetype = "AbsIndY"

    def visitPointerZ(self, node, env):
        node.nodetype = "AbsIndZ"

    def visitUnknown(self, node, env):
        pass


class Assembler(Pass):
    """Converts the IR into a list of bytes, suitable for writing to
    a file."""
    name = "Assembler"

    # The self.listing field defined in prePass and referred to elsewhere
    # holds the necessary information to build the program listing later.
    # Each member of this list is either a string (in which case it is
    # a listed instruction to be echoed) or it is a tuple of
    # (startPC, bytes). Byte listings need to be interrupted when the
    # PC is changed by an .org, which may happen in the middle of
    # data definition blocks.
    def prePass(self):
        self.output = []
        self.code = 0
        self.data = 0
        self.filler = 0
        if Cmd.listfile is not None:
            self.listing = Listing.Listing(Cmd.listfile)
        else:
            self.listing = Listing.NullLister()
        if Cmd.mapfile is not None:
            self.mapper = Listing.LabelMapper(Cmd.mapfile)
        else:
            self.mapper = Listing.NullLabelMapper()

    def postPass(self):
        self.listing.dump()
        self.mapper.dump()
        if Cmd.print_summary and Err.count == 0:
            print("Assembly complete: %s bytes output " \
                               "(%s code, %s data, %s filler)" \
                               % (len(self.output),
                                  self.code, self.data, self.filler), file=sys.stderr)

    def outputbyte(self, expr, env, tee=None):
        'Outputs a byte, with range checking'
        if self.writeOK:
            val = expr.value(env)
            if val < 0x00 or val > 0xff:
                Err.log("Byte constant " + str(expr) + " out of range")
                val = 0
            self.output.append(int(val))
            if tee is not None:
                tee.append(self.output[-1])
        else:
            Err.log("Attempt to write to data segment")

    def outputword(self, expr, env, tee=None):
        'Outputs a little-endian word, with range checking'
        if self.writeOK:
            val = expr.value(env)
            if val < 0x0000 or val > 0xFFFF:
                Err.log("Word constant " + str(expr) + " out of range")
                val = 0
            self.output.append(int(val & 0xFF))
            self.output.append(int((val >> 8) & 0xFF))
            if tee is not None:
                tee.extend(self.output[-2:])
        else:
            Err.log("Attempt to write to data segment")

    def outputdword(self, expr, env, tee=None):
        'Outputs a little-endian dword, with range checking'
        if self.writeOK:
            val = expr.value(env)
            if val < 0x00000000 or val > 0xFFFFFFFF:
                Err.log("DWord constant " + str(expr) + " out of range")
                val = 0
            self.output.append(int(val & 0xFF))
            self.output.append(int((val >> 8) & 0xFF))
            self.output.append(int((val >> 16) & 0xFF))
            self.output.append(int((val >> 24) & 0xFF))
            if tee is not None:
                tee.extend(self.output[-4:])
        else:
            Err.log("Attempt to write to data segment")

    def outputword_be(self, expr, env, tee=None):
        'Outputs a big-endian word, with range checking'
        if self.writeOK:
            val = expr.value(env)
            if val < 0x0000 or val > 0xFFFF:
                Err.log("Word constant " + str(expr) + " out of range")
                val = 0
            self.output.append(int((val >> 8) & 0xFF))
            self.output.append(int(val & 0xFF))
            if tee is not None:
                tee.extend(self.output[-2:])
        else:
            Err.log("Attempt to write to data segment")

    def outputdword_be(self, expr, env, tee=None):
        'Outputs a big-endian dword, with range checking'
        if self.writeOK:
            val = expr.value(env)
            if val < 0x00000000 or val > 0xFFFFFFFF:
                Err.log("DWord constant " + str(expr) + " out of range")
                val = 0
            self.output.append(int((val >> 24) & 0xFF))
            self.output.append(int((val >> 16) & 0xFF))
            self.output.append(int((val >> 8) & 0xFF))
            self.output.append(int(val & 0xFF))
            if tee is not None:
                tee.extend(self.output[-4:])
        else:
            Err.log("Attempt to write to data segment")

    def relativize(self, expr, env, arglen):
        "Convert an expression into one for use in relative addressing"
        arg = expr.value(env)
        arg = arg - (env.getPC() + arglen + 1)
        if arg < -128 or arg > 127:
            Err.log("Branch target out of bounds")
            arg = 0
        if arg < 0:
            arg += 256
        return IR.ConstantExpr(arg)

    def relativizelong(self, expr, env, arglen):
        "Convert an expression into one for use in relative addressing"
        arg = expr.value(env)
        arg = arg - (env.getPC() + arglen)
        if arg < 0:
            arg += 65536
        return IR.ConstantExpr(arg)

    def listing_string(self, pc, binary, mode, opcode, val1, val2):
        base = " %04X " % pc
        base += (" %02X" * len(binary)) % tuple(binary)
        formats = ["",
                   "#$%02X",
                   "#$%04X",
                   "$%02X",
                   "$%02X, X",
                   "$%02X, Y",
                   "$%04X",
                   "$%04X, X",
                   "$%04X, Y",
                   "($%04X)",
                   "($%04X, X)",
                   "($%04X), Y",
                   "($%04X), Z",
                   "($%02X)",
                   "($%02X, X)",
                   "($%02X), Y",
                   "($%02X, SP), Y",
                   "($%02X), Z",
                   "$%04X",
                   "$%04X",
                   "$%02X, $%04X"]
        fmt = ("%-16s %-5s" % (base, opcode.upper())) + formats[mode]
        if val1 is None:
            return fmt
        elif val2 is None:
            arglen = Ops.lengths[mode]
            mask = 0xFF
            # Relative is a full address in a byte, so it also has the
            # 0xFFFF mask.
            if arglen == 2 or mode == Ops.modes.index("Relative"):
                mask = 0xFFFF
            return fmt % (val1 & mask)
        else:
            # Mode is "Zero Page, Relative"
            return fmt % (val1 & 0xFF, val2 & 0xFFFF)

    def assemble(self, node, mode, env):
        "A generic instruction called by the visitor methods themselves"
        (opcode, expr, expr2) = node.data
        bin_op = Ops.opcodes[opcode][mode]
        if bin_op is None:
            Err.log('%s does not have mode "%s"' % (opcode.upper(),
                                                    Ops.modes[mode]))
            return
        inst_bytes = []
        self.outputbyte(IR.ConstantExpr(bin_op), env, inst_bytes)
        arglen = Ops.lengths[mode]
        val1 = None
        val2 = None
        if expr is not None:
            val1 = expr.value(env)
        if expr2 is not None:
            val2 = expr2.value(env)
        if mode == Ops.modes.index("Zero Page, Relative"):
            expr2 = self.relativize(expr2, env, arglen)
            self.outputbyte(expr, env, inst_bytes)
            self.outputbyte(expr2, env, inst_bytes)
        else:
            if mode == Ops.modes.index("Relative"):
                expr = self.relativize(expr, env, arglen)
            elif mode == Ops.modes.index("RelativeLong"):
                expr = self.relativizelong(expr, env, arglen)
            if arglen == 1:
                self.outputbyte(expr, env, inst_bytes)
            elif arglen == 2:
                self.outputword(expr, env, inst_bytes)
        self.listing.listInstruction(self.listing_string(env.getPC(),
                                                         inst_bytes,
                                                         mode, opcode,
                                                         val1, val2))
        env.incPC(1 + arglen)
        self.code += 1 + arglen

    def visitImplied(self, node, env):
        self.assemble(node,  Ops.modes.index("Implied"), env)

    def visitImmediate(self, node, env):
        self.assemble(node,  Ops.modes.index("Immediate"), env)

    def visitImmediateLong(self, node, env):
        self.assemble(node,  Ops.modes.index("ImmediateLong"), env)

    def visitZeroPage(self, node, env):
        self.assemble(node,  Ops.modes.index("Zero Page"), env)

    def visitZeroPageX(self, node, env):
        self.assemble(node,  Ops.modes.index("Zero Page, X"), env)

    def visitZeroPageY(self, node, env):
        self.assemble(node,  Ops.modes.index("Zero Page, Y"), env)

    def visitAbsolute(self, node, env):
        self.assemble(node,  Ops.modes.index("Absolute"), env)

    def visitAbsoluteX(self, node, env):
        self.assemble(node,  Ops.modes.index("Absolute, X"), env)

    def visitAbsoluteY(self, node, env):
        self.assemble(node,  Ops.modes.index("Absolute, Y"), env)

    def visitIndirect(self, node, env):
        self.assemble(node,  Ops.modes.index("(Absolute)"), env)

    def visitAbsIndX(self, node, env):
        self.assemble(node,  Ops.modes.index("(Absolute, X)"), env)

    def visitAbsIndY(self, node, env):
        self.assemble(node, Ops.modes.index("(Absolute), Y"), env)

    def visitAbsIndZ(self, node, env):
        self.assemble(node, Ops.modes.index("(Absolute), Z"), env)

    def visitZPIndirect(self, node, env):
        self.assemble(node, Ops.modes.index("(Zero Page)"), env)

    def visitIndirectX(self, node, env):
        self.assemble(node, Ops.modes.index("(Zero Page, X)"), env)

    def visitIndirectY(self, node, env):
        self.assemble(node, Ops.modes.index("(Zero Page), Y"), env)

    def visitIndirectZ(self, node, env):
        self.assemble(node, Ops.modes.index("(Zero Page), Z"), env)

    def visitIndirectSPY(self, node, env):
        self.assemble(node, Ops.modes.index("(Zero Page, SP), Y"), env)

    def visitRelative(self, node, env):
        self.assemble(node, Ops.modes.index("Relative"), env)

    def visitRelativeLong(self, node, env):
        self.assemble(node, Ops.modes.index("RelativeLong"), env)

    def visitZPRelative(self, node, env):
        self.assemble(node, Ops.modes.index("Zero Page, Relative"), env)

    def visitLabel(self, node, env):
        (label, val) = node.data
        location = val.value(env)
        self.mapper.mapLabel(label, str(node.ppt), location)

    def visitByte(self, node, env):
        created = []
        for expr in node.data:
            self.outputbyte(expr, env, created)
        self.registerData(created, env.getPC())
        env.incPC(len(node.data))
        self.data += len(node.data)

    def visitByteRange(self, node, env):
        offset = node.data[0].value(env) + 2
        length = node.data[1].value(env)
        if offset < 2:
            Err.log("Negative offset in .incbin")
        elif offset > len(node.data):
            Err.log("Offset extends past end of file")
        elif length < 0:
            Err.log("Negative length")
        elif offset + length > len(node.data):
            Err.log("File too small for .incbin subrange")
        else:
            created = []
            for expr in node.data[offset:(offset + length)]:
                self.outputbyte(expr, env, created)
            self.registerData(created, env.getPC())
            env.incPC(length)
            self.data += length

    def visitWord(self, node, env):
        created = []
        for expr in node.data:
            self.outputword(expr, env, created)
        self.registerData(created, env.getPC())
        env.incPC(len(node.data) * 2)
        self.data += len(node.data) * 2

    def visitDword(self, node, env):
        created = []
        for expr in node.data:
            self.outputdword(expr, env, created)
        self.registerData(created, env.getPC())
        env.incPC(len(node.data) * 4)
        self.data += len(node.data) * 4

    def visitWordBE(self, node, env):
        created = []
        for expr in node.data:
            self.outputword_be(expr, env, created)
        self.registerData(created, env.getPC())
        env.incPC(len(node.data) * 2)
        self.data += len(node.data) * 2

    def visitDwordBE(self, node, env):
        created = []
        for expr in node.data:
            self.outputdword_be(expr, env, created)
        self.registerData(created, env.getPC())
        env.incPC(len(node.data) * 4)
        self.data += len(node.data) * 4

    def visitSetPC(self, node, env):
        val = node.data[0].value(env)
        if self.writeOK and env.getPC() != val:
            self.listing.listDivider(val)
        env.setPC(val)

    def visitCheckPC(self, node, env):
        pc = env.getPC()
        target = node.data[0].value(env)
        if (pc > target):
            Err.log(".checkpc assertion failed: $%x > $%x" % (pc, target))

    def visitAdvance(self, node, env):
        pc = env.getPC()
        target = node.data[0].value(env)
        if (pc > target):
            Err.log("Attempted to .advance backwards: $%x to $%x" %
                    (pc, target))
        else:
            created = []
            for i in range(target - pc):
                self.outputbyte(node.data[1], env, created)
            self.filler += target - pc
            self.registerData(created, env.getPC())
        env.setPC(target)

    def registerData(self, vals, pc):
        self.listing.listData(vals, pc)



"""Command line options data."""

import optparse

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

enable_branch_extend = True
enable_undoc_ops = False
enable_65c02_exts = False
enable_4502_exts = False

warn_on_branch_extend = True

print_summary = True
print_loaded_files = False
print_pass = False
print_ir = False
print_labels = False

infiles = None
outfile = None
listfile = None
mapfile = None


def parse_args(raw_args):
    "Populate the module's globals based on the command-line options given."
    global enable_collapse, enable_branch_extend
    global enable_undoc_ops, enable_65c02_exts, enable_4502_exts
    global warn_on_branch_extend
    global print_summary, print_loaded_files
    global print_pass, print_ir, print_labels
    global infiles, outfile, listfile, mapfile

    parser = optparse.OptionParser(
        usage="Usage: %prog [options] srcfile [srcfile ...]",
        version="Ophis 6502 cross-assembler, version 2.1")

    parser.add_option("-o", default=None, dest="outfile",
                      help="Output filename (default 'ophis.bin')")
    parser.add_option("-l", default=None, dest="listfile",
                      help="Listing filename (not created by default)")
    parser.add_option("-m", default=None, dest="mapfile",
                      help="Label-address map filename (not created by default)")

    ingrp = optparse.OptionGroup(parser, "Input options")
    ingrp.add_option("-u", "--undoc", action="store_true", default=False,
                     help="Enable 6502 undocumented opcodes")
    ingrp.add_option("-c", "--65c02", action="store_true", default=False,
                     dest="c02", help="Enable 65c02 extended instruction set")
    ingrp.add_option("-4", "--4502", action="store_true", default=False,
                     dest="csg4502", help="Enable 4502 extended instruction set")

    outgrp = optparse.OptionGroup(parser, "Console output options")
    outgrp.add_option("-v", "--verbose", action="store_const", const=2,
                      help="Verbose mode", default=1)
    outgrp.add_option("-q", "--quiet", action="store_const", help="Quiet mode",
                      dest="verbose", const=0)
    outgrp.add_option("-d", "--debug", action="count", dest="verbose",
                      help=optparse.SUPPRESS_HELP)
    outgrp.add_option("--no-warn", action="store_false", dest="warn",
                      default=True, help="Do not print warnings")

    bingrp = optparse.OptionGroup(parser, "Compilation options")
    bingrp.add_option("--no-branch-extend", action="store_false",
                      dest="enable_branch_extend", default="True",
                      help="Disable branch-extension pass")

    parser.add_option_group(ingrp)
    parser.add_option_group(outgrp)
    parser.add_option_group(bingrp)

    (options, args) = parser.parse_args(raw_args)

    if len(args) == 0:
        parser.error("No input files specified")
    if options.c02 and options.undoc:
        parser.error("--undoc and --65c02 are mutually exclusive")
    if options.c02 and options.csg4502:
        parser.error("--65c02 and --4502 are mutually exclusive")
    if options.csg4502 and options.undoc:
        parser.error("--undoc and --4502 are mutually exclusive")

    infiles = args
    outfile = options.outfile
    listfile = options.listfile
    mapfile = options.mapfile
    enable_branch_extend = options.enable_branch_extend
    enable_undoc_ops = options.undoc
    enable_65c02_exts = options.c02
    enable_4502_exts = options.csg4502
    warn_on_branch_extend = options.warn
    print_summary = options.verbose > 0  # no options set
    print_loaded_files = options.verbose > 1  # v
    print_pass = options.verbose > 2  # dd
    print_ir = options.verbose > 3  # ddd
    print_labels = options.verbose > 4  # dddd



"""Main controller routines for the Ophis assembler.

    When invoked as main, interprets its command line and goes from there.
    Otherwise, use run_ophis(cmdline-list) to use it inside a script."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

import sys
import os
import Ophis.Frontend
import Ophis.IR
import Ophis.CorePragmas
import Ophis.Passes
import Ophis.Errors as Err
import Ophis.Environment
import Ophis.CmdLine
import Ophis.Opcodes


def run_all():
    """Transforms the source infiles to a binary outfile.

    Returns a shell-style exit code: 1 if there were errors, 0 if there
    were no errors.

    """
    Err.count = 0
    z = Ophis.Frontend.parse(Ophis.CmdLine.infiles)
    env = Ophis.Environment.Environment()

    m = Ophis.Passes.ExpandMacros()
    i = Ophis.Passes.InitLabels()
    l_basic = Ophis.Passes.UpdateLabels()
    l = Ophis.Passes.FixPoint("label update", [l_basic],
                              lambda: not l_basic.changed)

    # The instruction selector is a bunch of fixpoints, and which
    # passes run depends on the command line options a bit.
    c_basic = Ophis.Passes.Collapse()
    c = Ophis.Passes.FixPoint("instruction selection 1", [l, c_basic],
                              lambda: not c_basic.changed)

    if Ophis.CmdLine.enable_branch_extend:
        b = Ophis.Passes.ExtendBranches()
        instruction_select = Ophis.Passes.FixPoint("instruction selection 2",
                                                   [c, b],
                                                   lambda: not b.changed)
    else:
        instruction_select = c
    a = Ophis.Passes.Assembler()

    passes = []
    passes.append(Ophis.Passes.DefineMacros())
    passes.append(Ophis.Passes.FixPoint("macro expansion", [m],
                                        lambda: not m.changed))
    passes.append(Ophis.Passes.FixPoint("label initialization", [i],
                                        lambda: not i.changed))
    passes.extend([Ophis.Passes.CircularityCheck(),
                   Ophis.Passes.CheckExprs(),
                   Ophis.Passes.EasyModes()])
    passes.append(instruction_select)
    passes.extend([Ophis.Passes.NormalizeModes(),
                   Ophis.Passes.UpdateLabels(),
                   a])

    for p in passes:
        p.go(z, env)

    if Err.count == 0:
        try:
            outfile = Ophis.CmdLine.outfile
            if outfile == '-':
                output = sys.stdout.buffer
                if sys.platform == "win32":
                    # We can't dump our binary in text mode; that would be
                    # disastrous. So, we'll do some platform-specific
                    # things here to force our stdout to binary mode.
                    import msvcrt
                    msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
            elif outfile is None:
                output = open('ophis.bin', 'wb')
            else:
                output = open(outfile, 'wb')
            output.write(bytes(a.output))
            output.flush()
            if outfile != '-':
                output.close()
            return 0
        except IOError:
            print("Could not write to " + outfile, file=sys.stderr)
            return 1
    else:
        Err.report()
        return 1


def run_ophis(args):
    Ophis.CmdLine.parse_args(args)
    Ophis.Frontend.pragma_modules.append(Ophis.CorePragmas)

    if Ophis.CmdLine.enable_undoc_ops:
        Ophis.Opcodes.opcodes.update(Ophis.Opcodes.undocops)
    elif Ophis.CmdLine.enable_65c02_exts:
        Ophis.Opcodes.opcodes.update(Ophis.Opcodes.c02extensions)
    elif Ophis.CmdLine.enable_4502_exts:
        Ophis.Opcodes.opcodes.update(Ophis.Opcodes.csg4502extensions)

    Ophis.CorePragmas.reset()
    return run_all()


if __name__ == '__main__':
    sys.exit(run_ophis(sys.argv[1:]))



"""Opcodes file.

    Tables for the assembly of 6502-family instructions, mapping
    opcodes and addressing modes to binary instructions."""

# Copyright 2002-2014 Michael C. Martin and additional contributors.
# You may use, modify, and distribute this file under the MIT
# license: See README for details.

# DO NOT EDIT THIS FILE DIRECTLY.
# This file was automatically generated by gensets.py based on the
# the tables in tools/opcodes. Edit those tables, not these.

# Names of addressing modes
modes = ["Implied",
         "Immediate",
         "ImmediateLong",
         "Zero Page",
         "Zero Page, X",
         "Zero Page, Y",
         "Absolute",
         "Absolute, X",
         "Absolute, Y",
         "(Absolute)",
         "(Absolute, X)",
         "(Absolute), Y",
         "(Absolute), Z",
         "(Zero Page)",
         "(Zero Page, X)",
         "(Zero Page), Y",
         "(Zero Page, SP), Y",
         "(Zero Page), Z",
         "Relative",
         "RelativeLong",
         "Zero Page, Relative"]


# Lengths of the argument
lengths = [0, 1, 2, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 2, 2]

opcodes = {
           'adc': [None, 0x69, None, 0x65, 0x75, None, 0x6D, 0x7D,
                   0x79, None, None, None, None, None, 0x61, 0x71, None, None, None, None, None],
           'adc.w': [None, None, None, None, None, None, 0x6D, 0x7D,
                     0x79, None, None, None, None, None, None, None, None, None, None, None, None],
           'and': [None, 0x29, None, 0x25, 0x35, None, 0x2D, 0x3D,
                   0x39, None, None, None, None, None, 0x21, 0x31, None, None, None, None, None],
           'and.w': [None, None, None, None, None, None, 0x2D, 0x3D,
                     0x39, None, None, None, None, None, None, None, None, None, None, None, None],
           'asl': [0x0A, None, None, 0x06, 0x16, None, 0x0E, 0x1E,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'asl.w': [None, None, None, None, None, None, 0x0E, 0x1E,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'bcc': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0x90, None, None],
           'bcs': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0xB0, None, None],
           'beq': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0xF0, None, None],
           'bit': [None, None, None, 0x24, None, None, 0x2C, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'bit.w': [None, None, None, None, None, None, 0x2C, None,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'bmi': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0x30, None, None],
           'bne': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0xD0, None, None],
           'bpl': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0x10, None, None],
           'brk': [0x00, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'bvc': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0x50, None, None],
           'bvs': [None, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, 0x70, None, None],
           'clc': [0x18, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'cld': [0xD8, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'cli': [0x58, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'clv': [0xB8, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'cmp': [None, 0xC9, None, 0xC5, 0xD5, None, 0xCD, 0xDD,
                   0xD9, None, None, None, None, None, 0xC1, 0xD1, None, None, None, None, None],
           'cmp.w': [None, None, None, None, None, None, 0xCD, 0xDD,
                     0xD9, None, None, None, None, None, None, None, None, None, None, None, None],
           'cpx': [None, 0xE0, None, 0xE4, None, None, 0xEC, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'cpx.w': [None, None, None, None, None, None, 0xEC, None,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'cpy': [None, 0xC0, None, 0xC4, None, None, 0xCC, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'cpy.w': [None, None, None, None, None, None, 0xCC, None,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'dec': [None, None, None, 0xC6, 0xD6, None, 0xCE, 0xDE,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'dec.w': [None, None, None, None, None, None, 0xCE, 0xDE,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'dex': [0xCA, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'dey': [0x88, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'eor': [None, 0x49, None, 0x45, 0x55, None, 0x4D, 0x5D,
                   0x59, None, None, None, None, None, 0x41, 0x51, None, None, None, None, None],
           'eor.w': [None, None, None, None, None, None, 0x4D, 0x5D,
                     0x59, None, None, None, None, None, None, None, None, None, None, None, None],
           'inc': [None, None, None, 0xE6, 0xF6, None, 0xEE, 0xFE,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'inc.w': [None, None, None, None, None, None, 0xEE, 0xFE,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'inx': [0xE8, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'iny': [0xC8, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'jmp': [None, None, None, None, None, None, 0x4C, None,
                   None, 0x6C, None, None, None, None, None, None, None, None, None, None, None],
           'jsr': [None, None, None, None, None, None, 0x20, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'lda': [None, 0xA9, None, 0xA5, 0xB5, None, 0xAD, 0xBD,
                   0xB9, None, None, None, None, None, 0xA1, 0xB1, None, None, None, None, None],
           'lda.w': [None, None, None, None, None, None, 0xAD, 0xBD,
                     0xB9, None, None, None, None, None, None, None, None, None, None, None, None],
           'ldx': [None, 0xA2, None, 0xA6, None, 0xB6, 0xAE, None,
                   0xBE, None, None, None, None, None, None, None, None, None, None, None, None],
           'ldx.w': [None, None, None, None, None, None, 0xAE, None,
                     0xBE, None, None, None, None, None, None, None, None, None, None, None, None],
           'ldy': [None, 0xA0, None, 0xA4, 0xB4, None, 0xAC, 0xBC,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'ldy.w': [None, None, None, None, None, None, 0xAC, 0xBC,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'lsr': [0x4A, None, None, 0x46, 0x56, None, 0x4E, 0x5E,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'lsr.w': [None, None, None, None, None, None, 0x4E, 0x5E,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'nop': [0xEA, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'ora': [None, 0x09, None, 0x05, 0x15, None, 0x0D, 0x1D,
                   0x19, None, None, None, None, None, 0x01, 0x11, None, None, None, None, None],
           'ora.w': [None, None, None, None, None, None, 0x0D, 0x1D,
                     0x19, None, None, None, None, None, None, None, None, None, None, None, None],
           'pha': [0x48, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'php': [0x08, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'pla': [0x68, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'plp': [0x28, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'rol': [0x2A, None, None, 0x26, 0x36, None, 0x2E, 0x3E,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'rol.w': [None, None, None, None, None, None, 0x2E, 0x3E,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'ror': [0x6A, None, None, 0x66, 0x76, None, 0x6E, 0x7E,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'ror.w': [None, None, None, None, None, None, 0x6E, 0x7E,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'rti': [0x40, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'rts': [0x60, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'sbc': [None, 0xE9, None, 0xE5, 0xF5, None, 0xED, 0xFD,
                   0xF9, None, None, None, None, None, 0xE1, 0xF1, None, None, None, None, None],
           'sbc.w': [None, None, None, None, None, None, 0xED, 0xFD,
                     0xF9, None, None, None, None, None, None, None, None, None, None, None, None],
           'sec': [0x38, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'sed': [0xF8, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'sei': [0x78, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'sta': [None, None, None, 0x85, 0x95, None, 0x8D, 0x9D,
                   0x99, None, None, None, None, None, 0x81, 0x91, None, None, None, None, None],
           'sta.w': [None, None, None, None, None, None, 0x8D, 0x9D,
                     0x99, None, None, None, None, None, None, None, None, None, None, None, None],
           'stx': [None, None, None, 0x86, None, 0x96, 0x8E, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'stx.w': [None, None, None, None, None, None, 0x8E, None,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'sty': [None, None, None, 0x84, 0x94, None, 0x8C, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'sty.w': [None, None, None, None, None, None, 0x8C, None,
                     None, None, None, None, None, None, None, None, None, None, None, None, None],
           'tax': [0xAA, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'tay': [0xA8, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'tsx': [0xBA, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'txa': [0x8A, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'txs': [0x9A, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
           'tya': [0x98, None, None, None, None, None, None, None,
                   None, None, None, None, None, None, None, None, None, None, None, None, None],
          }
undocops = {
            'anc': [None, 0x0B, None, None, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'ane': [None, 0x8B, None, None, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'arr': [None, 0x6B, None, None, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'asr': [None, 0x4B, None, None, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'dcp': [None, None, None, 0xC7, 0xD7, None, 0xCF, 0xDF,
                    0xDB, None, None, None, None, None, 0xC3, 0xD3, None, None, None, None, None],
            'isb': [None, None, None, 0xE7, 0xF7, None, 0xEF, 0xFF,
                    0xFB, None, None, None, None, None, 0xE3, 0xF3, None, None, None, None, None],
            'las': [None, None, None, None, None, None, None, None,
                    0xBB, None, None, None, None, None, None, None, None, None, None, None, None],
            'lax': [None, None, None, 0xA7, None, 0xB7, 0xAF, None,
                    0xBF, None, None, None, None, None, 0xA3, 0xB3, None, None, None, None, None],
            'lxa': [None, 0xAB, None, None, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'nop': [0xEA, None, None, 0x04, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'rla': [None, None, None, 0x27, 0x37, None, 0x2F, 0x3F,
                    0x3B, None, None, None, None, None, 0x23, 0x33, None, None, None, None, None],
            'rra': [None, None, None, 0x67, 0x77, None, 0x6F, 0x7F,
                    0x7B, None, None, None, None, None, 0x63, 0x73, None, None, None, None, None],
            'sax': [None, None, None, 0x87, None, 0x97, 0x8F, None,
                    None, None, None, None, None, None, 0x83, None, None, None, None, None, None],
            'sbx': [None, 0xCB, None, None, None, None, None, None,
                    None, None, None, None, None, None, None, None, None, None, None, None, None],
            'sha': [None, None, None, None, None, None, None, None,
                    0x9F, None, None, None, None, None, None, 0x93, None, None, None, None, None],
            'shs': [None, None, None, None, None, None, None, None,
                    0x9B, None, None, None, None, None, None, None, None, None, None, None, None],
            'shx': [None, None, None, None, None, None, None, None,
                    0x9E, None, None, None, None, None, None, None, None, None, None, None, None],
            'slo': [None, None, None, 0x07, 0x17, None, 0x0F, 0x1F,
                    0x1B, None, None, None, None, None, 0x03, 0x13, None, None, None, None, None],
            'sre': [None, None, None, 0x47, 0x57, None, 0x4F, 0x5F,
                    0x5B, None, None, None, None, None, 0x43, 0x53, None, None, None, None, None],
           }
c02extensions = {
                 'adc': [None, 0x69, None, 0x65, 0x75, None, 0x6D, 0x7D,
                         0x79, None, None, None, None, 0x72, 0x61, 0x71, None, None, None, None, None],
                 'and': [None, 0x29, None, 0x25, 0x35, None, 0x2D, 0x3D,
                         0x39, None, None, None, None, 0x32, 0x21, 0x31, None, None, None, None, None],
                 'bbr0': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x0F],
                 'bbr1': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x1F],
                 'bbr2': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x2F],
                 'bbr3': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x3F],
                 'bbr4': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x4F],
                 'bbr5': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x5F],
                 'bbr6': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x6F],
                 'bbr7': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x7F],
                 'bbs0': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x8F],
                 'bbs1': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0x9F],
                 'bbs2': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0xAF],
                 'bbs3': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0xBF],
                 'bbs4': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0xCF],
                 'bbs5': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0xDF],
                 'bbs6': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0xEF],
                 'bbs7': [None, None, None, None, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, 0xFF],
                 'bit': [None, 0x89, None, 0x24, 0x34, None, 0x2C, 0x3C,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'bra': [None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, 0x80, None, None],
                 'cmp': [None, 0xC9, None, 0xC5, 0xD5, None, 0xCD, 0xDD,
                         0xD9, None, None, None, None, 0xD2, 0xC1, 0xD1, None, None, None, None, None],
                 'dea': [0x3A, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'dec': [0x3A, None, None, 0xC6, 0xD6, None, 0xCE, 0xDE,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'eor': [None, 0x49, None, 0x45, 0x55, None, 0x4D, 0x5D,
                         0x59, None, None, None, None, 0x52, 0x41, 0x51, None, None, None, None, None],
                 'ina': [0x1A, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'inc': [0x1A, None, None, 0xE6, 0xF6, None, 0xEE, 0xFE,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'jmp': [None, None, None, None, None, None, 0x4C, None,
                         None, 0x6C, 0x7C, None, None, None, None, None, None, None, None, None, None],
                 'lda': [None, 0xA9, None, 0xA5, 0xB5, None, 0xAD, 0xBD,
                         0xB9, None, None, None, None, 0xB2, 0xA1, 0xB1, None, None, None, None, None],
                 'ora': [None, 0x09, None, 0x05, 0x15, None, 0x0D, 0x1D,
                         0x19, None, None, None, None, 0x12, 0x01, 0x11, None, None, None, None, None],
                 'phx': [0xDA, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'phy': [0x5A, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'plx': [0xFA, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'ply': [0x7A, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb0': [None, None, None, 0x07, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb1': [None, None, None, 0x17, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb2': [None, None, None, 0x27, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb3': [None, None, None, 0x37, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb4': [None, None, None, 0x47, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb5': [None, None, None, 0x57, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb6': [None, None, None, 0x67, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'rmb7': [None, None, None, 0x77, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'sbc': [None, 0xE9, None, 0xE5, 0xF5, None, 0xED, 0xFD,
                         0xF9, None, None, None, None, 0xF2, 0xE1, 0xF1, None, None, None, None, None],
                 'smb0': [None, None, None, 0x87, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb1': [None, None, None, 0x97, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb2': [None, None, None, 0xA7, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb3': [None, None, None, 0xB7, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb4': [None, None, None, 0xC7, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb5': [None, None, None, 0xD7, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb6': [None, None, None, 0xE7, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'smb7': [None, None, None, 0xF7, None, None, None, None,
                          None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'sta': [None, None, None, 0x85, 0x95, None, 0x8D, 0x9D,
                         0x99, None, None, None, None, 0x92, 0x81, 0x91, None, None, None, None, None],
                 'stp': [0xDB, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'stz': [None, None, None, 0x64, 0x74, None, 0x9C, 0x9E,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'stz.w': [None, None, None, None, None, None, 0x9C, 0x9E,
                           None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'trb': [None, None, None, 0x14, None, None, 0x1C, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'trb.w': [None, None, None, None, None, None, 0x1C, None,
                           None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'tsb': [None, None, None, 0x04, None, None, 0x0C, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'tsb.w': [None, None, None, None, None, None, 0x0C, None,
                           None, None, None, None, None, None, None, None, None, None, None, None, None],
                 'wai': [0xCB, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None, None, None, None, None, None],
                }
csg4502extensions = {
                     'adc': [None, 0x69, None, 0x65, 0x75, None, 0x6D, 0x7D,
                             0x79, None, None, None, None, None, 0x61, 0x71, None, 0x72, None, None, None],
                     'and': [None, 0x29, None, 0x25, 0x35, None, 0x2D, 0x3D,
                             0x39, None, None, None, None, None, 0x21, 0x31, None, 0x32, None, None, None],
                     'asr': [0x43, None, None, None, 0x54, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'asw': [None, None, None, None, None, None, 0xCB, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'bbr0': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x0F],
                     'bbr1': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x1F],
                     'bbr2': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x2F],
                     'bbr3': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x3F],
                     'bbr4': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x4F],
                     'bbr5': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x5F],
                     'bbr6': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x6F],
                     'bbr7': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x7F],
                     'bbs0': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x8F],
                     'bbs1': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0x9F],
                     'bbs2': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0xAF],
                     'bbs3': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0xBF],
                     'bbs4': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0xCF],
                     'bbs5': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0xDF],
                     'bbs6': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0xEF],
                     'bbs7': [None, None, None, None, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, 0xFF],
                     'bcc': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0x90, 0x93, None],
                     'bcs': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0xB0, 0xB3, None],
                     'beq': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0xF0, 0xF3, None],
                     'bit': [None, 0x89, None, 0x24, 0x34, None, 0x2C, 0x3C,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'bmi': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0x30, 0x33, None],
                     'bne': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0xD0, 0xD3, None],
                     'bpl': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0x10, 0x13, None],
                     'bra': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0x80, 0x83, None],
                     'bsr': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, 0x63, None],
                     'bvc': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0x50, 0x53, None],
                     'bvs': [None, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, 0x70, 0x73, None],
                     'cle': [0x02, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'cmp': [None, 0xC9, None, 0xC5, 0xD5, None, 0xCD, 0xDD,
                             0xD9, None, None, None, None, None, 0xC1, 0xD1, None, 0xD2, None, None, None],
                     'cpz': [None, 0xC2, None, 0xD4, None, None, 0xDC, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'dea': [0x3A, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'dec': [0x3A, None, None, 0xC6, 0xD6, None, 0xCE, 0xDE,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'dew': [None, None, None, 0xC3, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'dez': [0x3B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'eom': [0xEA, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'eor': [None, 0x49, None, 0x45, 0x55, None, 0x4D, 0x5D,
                             0x59, None, None, None, None, None, 0x41, 0x51, None, 0x52, None, None, None],
                     'inc': [0x1A, None, None, 0xE6, 0xF6, None, 0xEE, 0xFE,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'inw': [None, None, None, 0xE3, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'inz': [0x1B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'jmp': [None, None, None, None, None, None, 0x4C, None,
                             None, 0x6C, 0x7C, None, None, None, None, None, None, None, None, None, None],
                     'jsr': [None, None, None, None, None, None, 0x20, None,
                             None, 0x22, 0x23, None, None, None, None, None, None, None, None, None, None],
                     'lda': [None, 0xA9, None, 0xA5, 0xB5, None, 0xAD, 0xBD,
                             0xB9, None, None, None, None, None, 0xA1, 0xB1, 0xE2, 0xB2, None, None, None],
                     'ldz': [None, 0xA3, None, None, None, None, 0xAB, 0xBB,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'map': [0x5C, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'neg': [0x42, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'ora': [None, 0x09, None, 0x05, 0x15, None, 0x0D, 0x1D,
                             0x19, None, None, None, None, None, 0x01, 0x11, None, 0x12, None, None, None],
                     'phw': [None, None, 0xF4, None, None, None, 0xFC, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'phx': [0xDA, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'phy': [0x5A, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'phz': [0xDB, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'plx': [0xFA, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'ply': [0x7A, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'plz': [0xFB, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb0': [None, None, None, 0x07, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb1': [None, None, None, 0x17, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb2': [None, None, None, 0x27, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb3': [None, None, None, 0x37, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb4': [None, None, None, 0x47, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb5': [None, None, None, 0x57, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb6': [None, None, None, 0x67, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rmb7': [None, None, None, 0x77, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'row': [None, None, None, None, None, None, 0xEB, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'rts': [0x60, 0x62, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'sbc': [None, 0xE9, None, 0xE5, 0xF5, None, 0xED, 0xFD,
                             0xF9, None, None, None, None, None, 0xE1, 0xF1, None, 0xF2, None, None, None],
                     'see': [0x03, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb0': [None, None, None, 0x87, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb1': [None, None, None, 0x97, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb2': [None, None, None, 0xA7, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb3': [None, None, None, 0xB7, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb4': [None, None, None, 0xC7, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb5': [None, None, None, 0xD7, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb6': [None, None, None, 0xE7, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'smb7': [None, None, None, 0xF7, None, None, None, None,
                              None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'sta': [None, None, None, 0x85, 0x95, None, 0x8D, 0x9D,
                             0x99, None, None, None, None, None, 0x81, 0x91, 0x82, 0x92, None, None, None],
                     'stx': [None, None, None, 0x86, None, 0x96, 0x8E, None,
                             0x9B, None, None, None, None, None, None, None, None, None, None, None, None],
                     'sty': [None, None, None, 0x84, 0x94, None, 0x8C, 0x8B,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'stz': [None, None, None, 0x64, 0x74, None, 0x9C, 0x9E,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'tab': [0x5B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'taz': [0x4B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'tba': [0x7B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'trb': [None, None, None, 0x14, None, None, 0x1C, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'tsb': [None, None, None, 0x04, None, None, 0x0C, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'tsy': [0x0B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'tys': [0x2B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                     'tza': [0x6B, None, None, None, None, None, None, None,
                             None, None, None, None, None, None, None, None, None, None, None, None, None],
                    }
